module SelectList exposing
    ( SelectList(..), fromLists, fromList, singleton
    , toList, selected, listBefore, listAfter, index, isHead, isLast
    , length, afterLength, beforeLength
    , attempt, append, prepend
    , Direction(..)
    , modify, set, insert, delete
    , changePosition, changePositionToEnd, moveSelection, moveSelectionToEnd
    , select, selectAll
    , map, Position(..), mapBy, mapBy_
    )

{-| Yet another SelectList implementation

A SelectList is a non-empty list which always has exactly one element selected.
It is an example of a list zipper.

Inspired by these modules

  - [rtfeldman/selectlist](http://package.elm-lang.org/packages/rtfeldman/selectlist/latest)
  - [turboMaCk/lazy-tree-with-zipper](http://package.elm-lang.org/packages/turboMaCk/lazy-tree-with-zipper/latest)

[`mapBy`](#mapBy) is the main function in this package.
Use [`mapBy`](#mapBy) in view.

    view : SelectList String -> Html Msg
    view selectList =
        ul [] <|
            SelectList.mapBy
                (\position item ->
                    li [ onClick (Set item) ]
                        [ text <| toString <| SelectList.index item
                        , toString <| SelectList.selected item
                        ]
                )
                selectList


# Type

@docs SelectList, fromLists, fromList, singleton


# Query

@docs toList, selected, listBefore, listAfter, index, isHead, isLast
@docs length, afterLength, beforeLength


# Operations

@docs attempt, append, prepend
@docs Direction
@docs modify, set, insert, delete


# Select

@docs changePosition, changePositionToEnd, moveSelection, moveSelectionToEnd
@docs select, selectAll


# Transformations

@docs map, Position, mapBy, mapBy_

-}


{-| A nonempty list which always has exactly one element selected.

Create one using `fromLists`, `fromList` or `singleton`.

-}
type SelectList a
    = SelectList (List a) a (List a)



-- constructor


{-| A `SelectList` if list has elements.

If empty, `Nothing`.

    fromList [] == Nothing

    fromList [ 2, 3, 4 ] == Just (fromLists [] 2 [ 3, 4 ])

-}
fromList : List a -> Maybe (SelectList a)
fromList list =
    case list of
        x :: xs ->
            Just <| SelectList [] x xs

        [] ->
            Nothing


{-| A `SelectList`.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> selected
        == 4

-}
fromLists : List a -> a -> List a -> SelectList a
fromLists before a after =
    SelectList (List.reverse before) a after


{-| A `SelectList` containing exactly one element.

    singleton 3 == fromLists [] 3 []

-}
singleton : a -> SelectList a
singleton a =
    SelectList [] a []


{-| Return a `List` containing the elements in a `SelectList`.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> toList
        == [ 1, 2, 3, 4, 5, 6 ]

-}
toList : SelectList a -> List a
toList (SelectList before a after) =
    List.reverse before ++ a :: after



-- Query


{-| Return the selected element.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> selected
        == 4

-}
selected : SelectList a -> a
selected (SelectList _ a _) =
    a


{-| Return the elements before the selected element.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> listBefore
        == [ 1, 2, 3 ]

-}
listBefore : SelectList a -> List a
listBefore (SelectList xs _ _) =
    List.reverse xs


{-| Return the elements after the selected element.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> listAfter
        == [ 5, 6 ]

-}
listAfter : SelectList a -> List a
listAfter (SelectList _ _ xs) =
    xs


{-| Check if the selected element is first element.

    fromLists [] 4 [ 5, 6 ]
        |> isHead
        == True

-}
isHead : SelectList a -> Bool
isHead (SelectList before _ _) =
    List.isEmpty before


{-| Check if the selected element is last element.

    fromLists [ 1, 2, 3 ] 4 []
        |> isLast
        == True

-}
isLast : SelectList a -> Bool
isLast (SelectList _ _ after) =
    List.isEmpty after


{-| Index of the selected element.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> index
        == 3

-}
index : SelectList a -> Int
index =
    beforeLength


{-| Length of `SelectList`.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> length
        == 6

-}
length : SelectList a -> Int
length (SelectList before _ after) =
    List.length before + 1 + List.length after


{-| Length of the elements before the selected element

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> beforeLength
        == 3

-}
beforeLength : SelectList a -> Int
beforeLength (SelectList before _ _) =
    List.length before


{-| Length of the elements after the selected element

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> afterLength
        == 2

-}
afterLength : SelectList a -> Int
afterLength (SelectList _ _ after) =
    List.length after



-- Operations


{-| Attempt to perform action over selectList and return original `SelectList`
in cases where this action returns `Nothing`.

    attempt f selectList =
        f selectList
            |> Maybe.withDefault selectList

-}
attempt : (SelectList a -> Maybe (SelectList a)) -> SelectList a -> SelectList a
attempt action selectList =
    Maybe.withDefault selectList <| action selectList


{-| Add elements to the end of a `SelectList`.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> append [ 7, 8 ]
        |> toList
        == [ 1, 2, 3, 4, 5, 6, 7, 8 ]

-}
append : List a -> SelectList a -> SelectList a
append list (SelectList before a after) =
    SelectList before a (after ++ list)


{-| Add elements to the beginning of a `SelectList`.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> prepend [ 7, 8 ]
        |> toList
        == [ 7, 8, 1, 2, 3, 4, 5, 6 ]

-}
prepend : List a -> SelectList a -> SelectList a
prepend list (SelectList before a after) =
    SelectList (before ++ List.reverse list) a after


{-| `Direction`
-}
type Direction
    = After
    | Before


{-| A selected element moves forward/backward.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> moveSelection Before 2
        == Just (fromLists [ 1 ] 4 [ 2, 3, 5, 6 ])

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> moveSelection After 3
        == Nothing

-}
moveSelection : Direction -> Int -> SelectList a -> Maybe (SelectList a)
moveSelection dir n selectList =
    if n < 0 then
        Nothing

    else if n == 0 then
        Just selectList

    else
        moveSelectionHelp dir selectList
            |> Maybe.andThen (moveSelection dir (n - 1))


moveSelectionHelp : Direction -> SelectList a -> Maybe (SelectList a)
moveSelectionHelp dir (SelectList before a after) =
    case dir of
        After ->
            case after of
                x :: xs ->
                    Just <| SelectList (x :: before) a xs

                [] ->
                    Nothing

        Before ->
            case before of
                x :: xs ->
                    Just <| SelectList xs a (x :: after)

                [] ->
                    Nothing


{-| A selected element moves to head/last.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> moveSelectionToEnd Before
        == Just (fromLists [] 4 [ 1, 2, 3, 5, 6 ])

    fromLists [ 1, 2, 3, 5, 6 ] 4 []
        |> moveSelectionToEnd After
        == Nothing

-}
moveSelectionToEnd : Direction -> SelectList a -> SelectList a
moveSelectionToEnd dir selectList =
    let
        moveLength =
            case dir of
                After ->
                    afterLength selectList

                Before ->
                    beforeLength selectList
    in
    attempt (moveSelection dir moveLength) selectList


{-| A selectList selects a element before/after the selected element.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> changePosition After 2
        == Just (fromLists [ 1, 2, 3, 4, 5 ] 6 [])

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> changePosition After 3
        == Nothing

-}
changePosition : Direction -> Int -> SelectList a -> Maybe (SelectList a)
changePosition dir n selectList =
    if n < 0 then
        Nothing

    else if n == 0 then
        Just selectList

    else
        changePositionHelp dir selectList
            |> Maybe.andThen (changePosition dir (n - 1))


changePositionHelp : Direction -> SelectList a -> Maybe (SelectList a)
changePositionHelp dir (SelectList before a after) =
    case dir of
        After ->
            case after of
                x :: xs ->
                    Just <| SelectList (a :: before) x xs

                [] ->
                    Nothing

        Before ->
            case before of
                x :: xs ->
                    Just <| SelectList xs x (a :: after)

                [] ->
                    Nothing


{-| A selectList selects a head/last element.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> changePositionToEnd After
        == fromLists [ 1, 2, 3, 4, 5 ] 6 []

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> changePositionToEnd Before
        == fromLists [] 1 [ 2, 3, 4, 5, 6 ]

-}
changePositionToEnd : Direction -> SelectList a -> SelectList a
changePositionToEnd dir selectList =
    let
        moveLength =
            case dir of
                After ->
                    afterLength selectList

                Before ->
                    beforeLength selectList
    in
    attempt (changePosition dir moveLength) selectList


{-| Modify the selected element using given function.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> modify (\selected -> 2 * selected)
        == fromLists [ 1, 2, 3 ] 8 [ 5, 6 ]

-}
modify : (a -> a) -> SelectList a -> SelectList a
modify f (SelectList before a after) =
    SelectList before (f a) after


{-| Replace the selected element with new one.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> set 10
        == fromLists [ 1, 2, 3 ] 10 [ 5, 6 ]

-}
set : a -> SelectList a -> SelectList a
set a (SelectList before _ after) =
    SelectList before a after


{-| Delete the selected element, then select `After`/`Before` element.

Returns Nothing if list of `After`/`Before` elements is empty.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> delete After
        == Just (fromLists [ 1, 2, 3 ] 5 [ 6 ])

    fromLists [ 1, 2, 3 ] 4 []
        |> delete After
        == Just (fromLists [ 1, 2 ] 3 [])

    fromLists [] 4 []
        |> delete After
        == Nothing

-}
delete : Direction -> SelectList a -> Maybe (SelectList a)
delete dir (SelectList before _ after) =
    case ( dir, before, after ) of
        ( _, [], [] ) ->
            Nothing

        ( _, x :: xs, [] ) ->
            Just <| SelectList xs x []

        ( _, [], x :: xs ) ->
            Just <| SelectList [] x xs

        ( After, _, x :: xs ) ->
            Just <| SelectList before x xs

        ( Before, x :: xs, _ ) ->
            Just <| SelectList xs x after


{-| Insert the new selected element, then move the old `After`/`Before`.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> insert After 8
        == fromLists [ 1, 2, 3 ] 8 [ 4, 5, 6 ]

-}
insert : Direction -> a -> SelectList a -> SelectList a
insert dir x (SelectList before a after) =
    case dir of
        After ->
            SelectList before x (a :: after)

        Before ->
            SelectList (a :: before) x after


{-| Change the selected element to the nearest one which passes a predicate function.
Find the list after selected element preferentially.

    isEven num =
        num % 2 == 0

    fromLists [ 1, 2 ] 3 [ 4, 5, 6 ]
        |> select isEven
        == fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]

-}
select : (a -> Bool) -> SelectList a -> Maybe (SelectList a)
select pred (SelectList before a after) =
    selectHelp pred before a after
        |> Maybe.map
            (\( nextBefore, next, nextAfter ) ->
                SelectList nextBefore next nextAfter
            )


selectHelp : (a -> Bool) -> List a -> a -> List a -> Maybe ( List a, a, List a )
selectHelp pred before a after =
    case selectAfterHelp pred before a after of
        Just selectList ->
            Just selectList

        Nothing ->
            selectBeforeHelp pred before a after


selectAfterHelp : (a -> Bool) -> List a -> a -> List a -> Maybe ( List a, a, List a )
selectAfterHelp pred before a after =
    if pred a then
        Just ( before, a, after )

    else
        case after of
            [] ->
                Nothing

            x :: xs ->
                selectAfterHelp pred (a :: before) x xs


selectBeforeHelp : (a -> Bool) -> List a -> a -> List a -> Maybe ( List a, a, List a )
selectBeforeHelp pred before a after =
    if pred a then
        Just ( before, a, after )

    else
        case before of
            [] ->
                Nothing

            x :: xs ->
                selectBeforeHelp pred xs x (a :: after)



-- Transformations


{-| List of all selectList.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        == [ fromLists [] 1 [ 2, 3, 4, 5, 6 ]
           , fromLists [ 1 ] 2 [ 3, 4, 5, 6 ]
           , fromLists [ 1, 2 ] 3 [ 4, 5, 6 ]
           , fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
           , fromLists [ 1, 2, 3, 4 ] 5 [ 6 ]
           , fromLists [ 1, 2, 3, 4, 5 ] 6 []
           ]

-}
selectAll : SelectList a -> List (SelectList a)
selectAll list =
    selectBefore list ++ list :: selectAfter list


selectBefore : SelectList a -> List (SelectList a)
selectBefore list =
    List.range 1 (beforeLength list)
        |> List.reverse
        |> List.filterMap (\n -> changePosition Before n list)


selectAfter : SelectList a -> List (SelectList a)
selectAfter list =
    List.range 1 (afterLength list)
        |> List.filterMap (\n -> changePosition After n list)


{-| Transform each element of the `SelectList`.

    fromLists [ 1, 2, 3 ] 4 [ 5, 6 ]
        |> map (\num -> num * 2)
        == fromLists [ 2, 4, 6 ] 8 [ 10, 12 ]

-}
map : (a -> b) -> SelectList a -> SelectList b
map f (SelectList before a after) =
    SelectList (List.map f before) (f a) (List.map f after)


{-| `Position` is used with [`mapBy`](#mapBy).

`Position` is Selected if the selected element,
BeforeSelected if an element before the selected element,
and AfterSelected if an element after the selected element.

-}
type Position
    = BeforeSelected
    | Selected
    | AfterSelected


{-| Transform each element of the `SelectList`.

The transform function receives a `Position`
and `SelectList` which selects a transformed element.

[`mapBy`](#mapBy) is the main function in this package.
Use [`mapBy`](#mapBy) in view.

    view : SelectList String -> Html Msg
    view selectList =
        ul [] <|
            SelectList.mapBy
                (\position item ->
                    li [ onClick (Set item) ]
                        [ text <| toString <| SelectList.index item
                        , toString <| SelectList.selected item
                        ]
                )
                selectList

If you can not use non-empty list, use [`mapBy_`](#mapBy_) that receives `List` instead of `SelectList`.

-}
mapBy : (Position -> SelectList a -> b) -> SelectList a -> List b
mapBy f list =
    let
        before =
            selectBefore list
                |> List.map (f BeforeSelected)

        transformed =
            f Selected list

        after =
            selectAfter list
                |> List.map (f AfterSelected)
    in
    before ++ transformed :: after


{-| This receives `List` instead of `SelectList`.
-}
mapBy_ : (SelectList a -> b) -> List a -> List b
mapBy_ f list =
    case fromList list of
        Just selectList ->
            mapBy (always f) selectList

        Nothing ->
            []
